from __future__ import annotations
from dataclasses import dataclass, field
import typing as t

from piccolo.columns.operators import (
    Equal,
    GreaterEqualThan,
    GreaterThan,
    In,
    LessEqualThan,
    LessThan,
    Like,
    ILike,
    NotEqual,
    NotIn,
    NotLike,
    Operator,
)
from piccolo.columns.combination import Where
from piccolo.custom_types import Iterable
from piccolo.querystring import QueryString
from piccolo.utils.warnings import colored_warning

if t.TYPE_CHECKING:
    from piccolo.table import Table  # noqa
    from .column_types import ForeignKey  # noqa


@dataclass
class ColumnMeta:
    """
    We store as many attributes in ColumnMeta as possible, to help avoid name
    clashes with user defined attributes.
    """

    # General attributes:
    null: bool = True
    primary: bool = False
    key: bool = False
    unique: bool = False

    # Set by the Table Metaclass:
    _name: t.Optional[str] = None
    _table: t.Optional[Table] = None

    # Used by Foreign Keys:
    call_chain: t.List["ForeignKey"] = field(default_factory=lambda: [])
    table_alias: t.Optional[str] = None

    @property
    def name(self) -> str:
        if not self._name:
            raise ValueError(
                "`_name` isn't defined - the Table Metaclass should set it."
            )
        return self._name

    @name.setter
    def name(self, value: str):
        self._name = value

    @property
    def table(self) -> Table:
        if not self._table:
            raise ValueError(
                "`_table` isn't defined - the Table Metaclass should set it."
            )
        return self._table

    @property
    def engine_type(self) -> str:
        engine = self.table._meta.db
        if engine:
            return engine.engine_type
        else:
            raise ValueError("The table has no engine defined.")

    def get_full_name(self, just_alias=False) -> str:
        """
        Returns the full column name, taking into account joins.
        """
        column_name = self.name

        if not self.call_chain:
            return f"{self.table._meta.tablename}.{column_name}"

        column_name = (
            "$".join([i._meta.name for i in self.call_chain])
            + f"${column_name}"
        )
        alias = f"{self.call_chain[-1]._meta.table_alias}.{self.name}"
        if just_alias:
            return alias
        else:
            return f'{alias} AS "{column_name}"'


class Column:

    value_type: t.Type = int

    def __init__(
        self,
        null: bool = True,
        primary: bool = False,
        key: bool = False,
        unique: bool = False,
    ) -> None:
        self._meta = ColumnMeta(
            null=null, primary=primary, key=key, unique=unique
        )

    def is_in(self, values: Iterable) -> Where:
        return Where(column=self, values=values, operator=In)

    def not_in(self, values: Iterable) -> Where:
        return Where(column=self, values=values, operator=NotIn)

    def like(self, value: str) -> Where:
        if "%" not in value:
            raise ValueError("% is required for like operators")
        return Where(column=self, value=value, operator=Like)

    def ilike(self, value: str) -> Where:
        if "%" not in value:
            raise ValueError("% is required for ilike operators")
        if self._meta.engine_type == "postgres":
            operator: t.Type[Operator] = ILike
        else:
            colored_warning(
                "SQLite doesn't support ILIKE currently, falling back to LIKE."
            )
            operator = Like
        return Where(column=self, value=value, operator=operator)

    def not_like(self, value: str) -> Where:
        if "%" not in value:
            raise ValueError("% is required for like operators")
        return Where(column=self, value=value, operator=NotLike)

    def __lt__(self, value) -> Where:
        return Where(column=self, value=value, operator=LessThan)

    def __le__(self, value) -> Where:
        return Where(column=self, value=value, operator=LessEqualThan)

    def __gt__(self, value) -> Where:
        return Where(column=self, value=value, operator=GreaterThan)

    def __ge__(self, value) -> Where:
        return Where(column=self, value=value, operator=GreaterEqualThan)

    def __eq__(self, value) -> Where:  # type: ignore
        return Where(column=self, value=value, operator=Equal)

    def __ne__(self, value) -> Where:  # type: ignore
        return Where(column=self, value=value, operator=NotEqual)

    def __hash__(self):
        return hash(self._meta.name)

    @property
    def querystring(self) -> QueryString:
        """
        Used when creating tables.
        """
        column_type = getattr(
            self, "column_type", self.__class__.__name__.upper()
        )
        query = f"{self._meta.name} {column_type}"
        if self._meta.primary:
            query += " PRIMARY"
        if self._meta.key:
            query += " KEY"
        if self._meta.unique:
            query += " UNIQUE"

        foreign_key_meta = getattr(self, "_foreign_key_meta", None)
        if foreign_key_meta:
            references = foreign_key_meta.references
            query += f" REFERENCES {references._meta.tablename}"

        return QueryString(query)

    def __str__(self):
        return self.querystring.__str__()



from __future__ import annotations

from piccolo.custom_types import Combinable
from piccolo.query.base import Query
from piccolo.query.mixins import WhereDelegate
from piccolo.querystring import QueryString


class DeletionError(Exception):
    pass


class Delete(Query):

    __slots__ = ("where_delegate", "force")

    def __init__(self, force=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.force = force

    def _setup_delegates(self):
        self.where_delegate = WhereDelegate()

    def where(self, where: Combinable) -> Delete:
        self.where_delegate.where(where)
        return self

    def _validate(self):
        """
        Don't let a deletion happen unless it has a where clause, or is
        explicitly forced.
        """
        if (not self.where_delegate._where) and (not self.force):
            classname = self.table.__name__
            raise DeletionError(
                "Do you really want to delete all the data from "
                f"{classname}? If so, use {classname}.delete(force=True)."
            )

    @property
    def querystring(self) -> QueryString:
        query = f"DELETE FROM {self.table._meta.tablename}"
        if self.where_delegate._where:
            query += " WHERE {}"
            return QueryString(query, self.where_delegate._where.querystring)
        else:
            return QueryString(query)

    def __str__(self) -> str:
        query = f"DELETE FROM {self.table._meta.tablename}"
        if self.where_delegate._where:
            query += f" WHERE {self.where_delegate._where.__str__()}"
        return query



from ..base import DBTestCase
from ..example_project.tables import Band, Concert


class TestSelect(DBTestCase):
    def test_query_all_columns(self):
        self.insert_row()

        response = Band.select().run_sync()
        print(f"response = {response}")

        self.assertDictEqual(
            response[0],
            {"id": 1, "name": "Pythonistas", "manager": 1, "popularity": 1000},
        )

    def test_query_some_columns(self):
        self.insert_row()

        response = Band.select().columns(Band.name).run_sync()
        print(f"response = {response}")

        self.assertDictEqual(response[0], {"name": "Pythonistas"})

    def test_where_like(self):
        self.insert_rows()

        response = (
            Band.select()
            .columns(Band.name)
            .where(Band.name.like("Python%"))
            .run_sync()
        )

        print(f"response = {response}")

        self.assertEqual(response, [{"name": "Pythonistas"}])

    def test_where_ilike(self):
        self.insert_rows()

        response = (
            Band.select()
            .columns(Band.name)
            .where(Band.name.ilike("python%"))
            .run_sync()
        )

        print(f"response = {response}")

        self.assertEqual(response, [{"name": "Pythonistas"}])

    def test_where_not_like(self):
        self.insert_rows()

        response = (
            Band.select()
            .columns(Band.name)
            .where(Band.name.not_like("Python%"))
            .order_by(Band.name)
            .run_sync()
        )

        print(f"response = {response}")

        self.assertEqual(
            response, [{"name": "CSharps"}, {"name": "Rustaceans"}]
        )

    def test_where_greater_than(self):
        self.insert_rows()

        response = (
            Band.select()
            .columns(Band.name)
            .where(Band.popularity > 1000)
            .run_sync()
        )

        print(f"response = {response}")

        self.assertEqual(response, [{"name": "Rustaceans"}])

    def test_where_greater_equal_than(self):
        self.insert_rows()

        response = (
            Band.select()
            .columns(Band.name)
            .where(Band.popularity >= 1000)
            .order_by(Band.name)
            .run_sync()
        )

        print(f"response = {response}")

        self.assertEqual(
            response, [{"name": "Pythonistas"}, {"name": "Rustaceans"}]
        )

    def test_where_less_than(self):
        self.insert_rows()

        response = (
            Band.select()
            .columns(Band.name)
            .where(Band.popularity < 1000)
            .run_sync()
        )

        print(f"response = {response}")

        self.assertEqual(response, [{"name": "CSharps"}])

    def test_where_less_equal_than(self):
        self.insert_rows()

        response = (
            Band.select()
            .columns(Band.name)
            .where(Band.popularity <= 1000)
            .run_sync()
        )

        print(f"response = {response}")

        self.assertEqual(
            response, [{"name": "Pythonistas"}, {"name": "CSharps"}]
        )

    def test_where_and(self):
        self.insert_rows()

        response = (
            Band.select()
            .columns(Band.name)
            .where((Band.popularity <= 1000) & (Band.name.like("Python%")))
            .run_sync()
        )

        print(f"response = {response}")

        self.assertEqual(response, [{"name": "Pythonistas"}])

    def test_where_or(self):
        self.insert_rows()

        response = (
            Band.select()
            .columns(Band.name)
            .where((Band.name == "Rustaceans") | (Band.name == "CSharps"))
            .order_by(Band.name)
            .run_sync()
        )

        print(f"response = {response}")

        self.assertEqual(
            response, [{"name": "CSharps"}, {"name": "Rustaceans"}]
        )

    def test_multiple_where(self):
        """
        Test that chaining multiple where clauses works results in an AND.
        """
        self.insert_rows()

        query = (
            Band.select()
            .columns(Band.name)
            .where(Band.name == "Rustaceans")
            .where(Band.manager == 2)
        )

        response = query.run_sync()

        print(f"response = {response}")

        self.assertEqual(response, [{"name": "Rustaceans"}])
        self.assertTrue("AND" in query.__str__())

    def test_complex_where(self):
        """
        Test a complex where clause - combining AND, and OR.
        """
        self.insert_rows()

        query = (
            Band.select()
            .columns(Band.name)
            .where(
                ((Band.popularity == 2000) & (Band.manager == 2))
                | ((Band.popularity == 10) & (Band.manager == 3))
            )
            .order_by(Band.name)
        )

        response = query.run_sync()

        print(f"response = {response}")

        self.assertEqual(
            response, [{"name": "CSharps"}, {"name": "Rustaceans"}]
        )

    def test_limit(self):
        self.insert_rows()

        response = (
            Band.select()
            .columns(Band.name)
            .order_by(Band.name)
            .limit(1)
            .run_sync()
        )

        print(f"response = {response}")

        self.assertEqual(response, [{"name": "CSharps"}])

    def test_first(self):
        self.insert_rows()

        response = (
            Band.select()
            .columns(Band.name)
            .order_by(Band.name)
            .first()
            .run_sync()
        )

        print(f"response = {response}")

        self.assertEqual(response, {"name": "CSharps"})

    def test_order_by_ascending(self):
        self.insert_rows()

        response = (
            Band.select()
            .columns(Band.name)
            .order_by(Band.name)
            .limit(1)
            .run_sync()
        )

        print(f"response = {response}")

        self.assertEqual(response, [{"name": "CSharps"}])

    def test_order_by_decending(self):
        self.insert_rows()

        response = (
            Band.select()
            .columns(Band.name)
            .order_by(Band.name, ascending=False)
            .limit(1)
            .run_sync()
        )

        print(f"response = {response}")

        self.assertEqual(response, [{"name": "Rustaceans"}])

    def test_count(self):
        self.insert_rows()

        response = Band.count().where(Band.name == "Pythonistas").run_sync()

        print(f"response = {response}")

        self.assertEqual(response, 1)

    def test_distinct(self):
        self.insert_rows()
        self.insert_rows()

        response = (
            Band.select()
            .columns(Band.name)
            .where(Band.name == "Pythonistas")
            .distinct()
            .run_sync()
        )

        self.assertTrue(response == [{"name": "Pythonistas"}])

    def test_call_chain(self):
        """
        Make sure the call chain lengths are the correct size.
        """
        self.assertEqual(len(Concert.band_1.name._meta.call_chain), 1)
        self.assertEqual(len(Concert.band_1.manager.name._meta.call_chain), 2)




"""
Populates a database with an example schema and data, and launches a shell
for interacting with the data using Piccolo.
"""
import sys

import click

from piccolo import table
from piccolo import columns
from piccolo.engine import SQLiteEngine, PostgresEngine


class Manager(table.Table):
    name = columns.Varchar(length=50)


class Band(table.Table):
    name = columns.Varchar(length=50)
    manager = columns.ForeignKey(references=Manager)
    popularity = columns.Integer()


class Venue(table.Table):
    name = columns.Varchar(length=100)
    capacity = columns.Integer()


class Concert(table.Table):
    band_1 = columns.ForeignKey(Band)
    band_2 = columns.ForeignKey(Band)
    venue = columns.ForeignKey(Venue)


TABLES = (Manager, Band, Venue, Concert)


def populate():
    """
    Drop then recreate the tables, and populate with data.
    """
    for _table in reversed(TABLES):
        try:
            if _table.table_exists().run_sync():
                _table.alter().drop_table().run_sync()
        except Exception as e:
            print(e)

    for _table in TABLES:
        try:
            _table.create().run_sync()
        except Exception as e:
            print(e)

    guido = Manager(name="Guido")
    guido.save().run_sync()

    pythonistas = Band(name="Pythonistas", manager=guido.id, popularity=1000)
    pythonistas.save().run_sync()

    graydon = Manager(name="Graydon")
    graydon.save().run_sync()

    rustaceans = Band(name="Rustaceans", manager=graydon.id, popularity=500)
    rustaceans.save().run_sync()

    venue = Venue(name="Amazing Venue", capacity=5000)
    venue.save().run_sync()

    concert = Concert(
        band_1=pythonistas.id, band_2=rustaceans.id, venue=venue.id
    )
    concert.save().run_sync()


@click.command(name="playground")
@click.option(
    "--engine",
    default="sqlite",
    help="Which database engine to use",
    type=click.Choice(["sqlite", "postgres"]),
)
@click.option("--user", default="piccolo", help="Postgres user")
@click.option("--password", default="piccolo", help="Postgres password")
@click.option(
    "--database", default="piccolo_playground", help="Postgres database"
)
@click.option("--host", default="localhost", help="Postgres host")
@click.option("--port", default=5432, help="Postgres port")
def playground(engine, user, password, database, host, port):
    """
    Creates a test database to play with.
    """
    try:
        import IPython
    except ImportError:
        print(
            "Install iPython using `pip install ipython==7.6.1` to use this "
            "feature."
        )
        sys.exit(1)

    if engine.upper() == "POSTGRES":
        db = PostgresEngine(
            {
                "host": host,
                "database": database,
                "user": user,
                "password": password,
                "port": port,
            }
        )
        for _table in TABLES:
            _table._meta.db = db
    else:
        db = SQLiteEngine()
        for _table in TABLES:
            _table._meta.db = db

    print("Tables:\n")

    for _table in TABLES:
        print(_table)
        print("\n")

    populate()
    IPython.embed()
